/*
 * Copyright (c) 2012 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Fri Mar 23 2012
 */
#include "acgenomix.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "genomix.h"


/* --- global data --------------------------------------------------------- */

/* client handles */
struct gnmx_handle *gnmx_clients = NULL;


/* --- local data ---------------------------------------------------------- */

/* rpaths */
static const char **rpaths = NULL;
static size_t nrpaths = 0;

static void	gnmx_requestcb(genom_client c, int rqstid, int error,
			uint32_t report, void *output, void *data);
static int	gnmx_module_cmd(ClientData data, Tcl_Interp *interp,
			int objc, Tcl_Obj *const objv[]);
static void	gnmx_delete_module_cmd(ClientData data);

void gnmx_event_proc(ClientData data, int mask);


/* --- gnmx_rpath ---------------------------------------------------------- */

int
gnmx_rpath(ClientData data, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
{
  const char *path;
  const char **p;
  size_t i;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "path");
    return TCL_ERROR;
  }

  path = Tcl_GetString(objv[1]);
  for(i=0; i<nrpaths; i++)
    if (!strcmp(rpaths[i], path)) return TCL_OK;

  p = realloc(rpaths, (nrpaths+1)*sizeof(*rpaths));
  if (!p) {
    Tcl_AppendResult(interp, strerror(errno), NULL);
    return TCL_ERROR;
  }
  rpaths = p;
  rpaths[nrpaths] = strdup(path);
  nrpaths++;

  Tcl_SetObjResult(interp, objv[1]);
  return TCL_OK;
}


/* --- gnmx_load ----------------------------------------------------------- */

int
gnmx_load(ClientData data, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
{
  char path[PATH_MAX];
  struct genom_client_info **info;
  struct gnmx_handle *h;
  const char *name;
  const char *module;
  Tcl_CmdInfo cmdinfo;
  Tcl_Obj *cmd;
  Tcl_Obj **largv;
  struct stat sb;
  char msg[256];
  size_t i;
  int largc;

  if (objc < 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "path args");
    return TCL_ERROR;
  }
  module = Tcl_GetString(objv[1]);

  h = malloc(sizeof(*h));
  if (!h) {
    Tcl_AppendResult(interp, strerror(errno), NULL);
    return TCL_ERROR;
  }
  h->c = NULL;
  h->path = NULL;

  /* try rpaths first */
  for(i = 0; i<nrpaths; i++) {
    /* try a .so */
    snprintf(path, sizeof(path), "%s/%s.so", rpaths[i], module);
    if (stat(path, &sb) || !S_ISREG(sb.st_mode)) {
      snprintf(msg, sizeof(msg), "infox {%s: %s} 1", path, strerror(errno));
      Tcl_Eval(interp, msg);

      /* try a raw file */
      snprintf(path, sizeof(path), "%s/%s", rpaths[i], module);
      if (stat(path, &sb) || !S_ISREG(sb.st_mode)) {
        snprintf(msg, sizeof(msg), "infox {%s: %s} 1", path, strerror(errno));
        Tcl_Eval(interp, msg);
        continue;
      }
    }
    break;
  }
  if (i>=nrpaths) {
    /* try a .so in genomix prefix */
    snprintf(path, sizeof(path),
             "%s/genom/client/c/%s.so", GENOMIX_LIBDIR, module);
    if (stat(path, &sb) || !S_ISREG(sb.st_mode)) {
      snprintf(msg, sizeof(msg), "infox {%s: %s} 1", path, strerror(errno));
      Tcl_Eval(interp, msg);

     /* try a raw file in genomix prefix */
      snprintf(path, sizeof(path),
               "%s/genom/client/c/%s", GENOMIX_LIBDIR, module);
      if (stat(path, &sb) || !S_ISREG(sb.st_mode)) {
        snprintf(msg, sizeof(msg), "infox {%s: %s} 1", path, strerror(errno));
        Tcl_Eval(interp, msg);

        /* system defaults as the last chance */
        snprintf(path, sizeof(path), "%s", module);
      }
    }
  }

  /* load shared object */
  h->dlh = dlopen(path, RTLD_NOW | RTLD_LOCAL);
  if (!h->dlh) {
    Tcl_AppendResult(interp, dlerror(), NULL);
    goto error;
  }
  h->path = strdup(path);

  /* retrieve entry point */
  info = dlsym(h->dlh, GENOM_CLIENT_DLSYM);
  if (!info) {
    Tcl_AppendResult(interp, dlerror(), NULL);
    goto error;
  }
  h->info = *info;
  if (h->info->protocol != genom_client_protocol) {
    char num[16];
    Tcl_AppendResult(interp, "client protocol mismatch, client ", NULL);
    snprintf(num, sizeof(num), "%d", h->info->protocol);
    Tcl_AppendResult(interp, num, " vs. genomix ", NULL);
    snprintf(num, sizeof(num), "%d", genom_client_protocol);
    Tcl_AppendResult(interp, num, NULL);
    goto error;
  }

  /* get args */
  if (Tcl_ListObjGetElements(interp, objv[2], &largc, &largv) != TCL_OK)
    goto error;
  {
    char *argv[largc];

    for (i = 0; i<largc; i++)
      argv[i] = Tcl_GetString(largv[i]);

    /* setup communication - protect against bad atexit() */
    gnmx_atexit_push(h->dlh);
    h->c = h->info->init(largc, argv);
    if (!h->c) {
      Tcl_AppendResult(interp, h->info->strerror(h->c), NULL);
      gnmx_atexit_pop();
      goto error;
    }
    gnmx_atexit_pop();
  }

  /* create command */
  gnmx_atexit_push(h->dlh);
  name = h->info->instance(h->c);
  gnmx_atexit_pop();
  cmd = Tcl_NewStringObj("::module::", -1);
  Tcl_AppendToObj(cmd, name, -1);
  Tcl_IncrRefCount(cmd);

  if (Tcl_GetCommandInfo(interp, Tcl_GetString(cmd), &cmdinfo)) {
    /* exists: reuse existing command */
    Tcl_AppendResult(interp, name, NULL);
    gnmx_atexit_push(h->dlh);
    h->info->fini(h->c);
    gnmx_atexit_pop();
    gnmx_atexit_invoke(h->dlh);
    dlclose(h->dlh);
    free(h);
    Tcl_DecrRefCount(cmd);
    return TCL_OK;
  }

  h->token = Tcl_CreateObjCommand(
    interp, Tcl_GetString(cmd), gnmx_module_cmd, h, gnmx_delete_module_cmd);
  Tcl_DecrRefCount(cmd);
  if (!h->token) goto error;

  h->interp = interp;
  gnmx_atexit_push(h->dlh);
  h->eventfd = h->info->eventfd(h->c);
  gnmx_atexit_pop();
  h->next = gnmx_clients;
  gnmx_clients = h;
  gnmx_watch(h);

  Tcl_AppendResult(interp, name, NULL);
  return TCL_OK;

error:
  if (h) {
    if (h->c) {
      gnmx_atexit_push(h->dlh);
      h->info->fini(h->c);
      gnmx_atexit_pop();
    }
    if (h->dlh) {
      gnmx_atexit_invoke(h->dlh);
      dlclose(h->dlh);
    }
    if (h->path) free(h->path);
    free(h);
  }
  return TCL_ERROR;
}


/* --- gnmx_module_cmd ----------------------------------------------------- */

static int
gnmx_module_cmd(ClientData data, Tcl_Interp *interp,
                int objc, Tcl_Obj *const objv[])
{
  enum cidx {
    cidx_path, cidx_info, cidx_send, cidx_clean, cidx_read
  };
  static const char *args[] = {
    [cidx_path] = "path", [cidx_info] = "info", [cidx_send] = "send",
    [cidx_clean] = "clean", [cidx_read] = "read", NULL
  };

  struct gnmx_handle *h = data;
  enum cidx i;
  int s, e;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 0, objv, "command ?args?");
    return TCL_ERROR;
  }

  e = Tcl_GetIndexFromObj(interp, objv[1], args, "subcommand", 0, (int *)&i);
  if (e != TCL_OK) return e;

  switch(i) {
    case cidx_path: {
      Tcl_AppendResult(interp, h->path?h->path:"<null>", NULL);
      return h->path ? TCL_OK : TCL_ERROR;
    }

    case cidx_info: {
      Tcl_AppendResult(interp, "{", NULL);
      Tcl_AppendResult(interp, "\"name\":\"", h->info->name, "\",", NULL);
      Tcl_AppendResult(interp, "\"instance\":\"",
                       h->info->instance(h->c), "\",", NULL);

      Tcl_AppendResult(interp, "\"services\":{", NULL);
      for(s=0; s<h->info->nservices; s++) {
        if (s) Tcl_AppendResult(interp, ",", NULL);

        Tcl_AppendResult(interp, "\"",
                         h->info->services[s].name, "\":{", NULL);
        Tcl_AppendResult(interp, "\"input\":",
                         h->info->services[s].input, ",", NULL);
        Tcl_AppendResult(interp, "\"output\":",
                         h->info->services[s].output, ",", NULL);
        Tcl_AppendResult(interp, "\"meta\":", h->info->services[s].meta, NULL);
        Tcl_AppendResult(interp, "}", NULL);
      }
      Tcl_AppendResult(interp, "},", NULL);

      Tcl_AppendResult(interp, "\"ports\":{", NULL);
      for(s=0; s<h->info->nports; s++) {
        if (s) Tcl_AppendResult(interp, ",", NULL);

        Tcl_AppendResult(interp, "\"", h->info->ports[s].name, "\":{", NULL);
        if (h->info->ports[s].simple)
          Tcl_AppendResult(interp, "\"simple\":1,", NULL);
        else
          Tcl_AppendResult(interp, "\"simple\":0,", NULL);
        Tcl_AppendResult(interp, "\"data\":",
                         h->info->ports[s].data, ",", NULL);
        Tcl_AppendResult(interp, "\"meta\":", h->info->ports[s].meta, NULL);
        Tcl_AppendResult(interp, "}", NULL);
      }
      Tcl_AppendResult(interp, "}", NULL);

      Tcl_AppendResult(interp, "}", NULL);
      break;
    }

    case cidx_send: {
      const char *rname, *rarg;
      void *input;
      char *e;
      int a, r;

      if (objc != 4) {
        Tcl_WrongNumArgs(interp, 1, objv, "service arg");
        return TCL_ERROR;
      }
      rname = Tcl_GetString(objv[2]);
      rarg = Tcl_GetString(objv[3]);

      for(r=0; r < h->info->nservices; r++)
        if (!strcmp(h->info->services[r].name, rname)) break;
      if (r >= h->info->nservices) {
        Tcl_AppendResult(interp, "unknown service ", rname, NULL);
        return TCL_ERROR;
      }

      input = malloc(h->info->services[r].input_size);
      if (!input) {
        Tcl_AppendResult(interp, "not enough memory", NULL);
        return TCL_ERROR;
      }
      gnmx_atexit_push(h->dlh);
      h->info->services[r].init_input(input);
      s = h->info->services[r].json_scan(input, rarg, &e);
      gnmx_atexit_pop();
      if (s) {
        char ectx[64];
        int n;

        Tcl_AppendResult(interp, "invalid json string at offset ", NULL);
        snprintf(ectx, sizeof(ectx), "%d", (int)(e-rarg));
        Tcl_AppendResult(interp, ectx, ": ", NULL);
        n = e-rarg > 15 ? 15 : e-rarg;
        snprintf(ectx, sizeof(ectx), "%.*s<here>%.15s", n, e-n, e);
        Tcl_AppendResult(interp, ectx, NULL);
        return TCL_ERROR;
      }

      gnmx_atexit_push(h->dlh);
      s = h->info->services[r].send(h->c, input,
                                    gnmx_requestcb, gnmx_requestcb, h, &a);
      gnmx_atexit_pop();
      if (s) {
        Tcl_AppendResult(interp, h->info->strerror(h->c), NULL);
        return TCL_ERROR;
      }

      Tcl_SetObjResult(interp, Tcl_NewIntObj(a));
      break;
    }

    case cidx_clean: {
      int a;

      if (objc != 3) {
        Tcl_WrongNumArgs(interp, 2, objv, "rqstid");
        return TCL_ERROR;
      }
      if (Tcl_GetIntFromObj(interp, objv[2], &a) != TCL_OK)
        return TCL_ERROR;

      gnmx_atexit_push(h->dlh);
      s = h->info->clean(h->c, a);
      gnmx_atexit_pop();
      if (s) {
        Tcl_AppendResult(interp, h->info->strerror(h->c), NULL);
        return TCL_ERROR;
      }
      break;
    }

    case cidx_read: {
      const char *port, *name;
      char *data;
      char *json;
      int p;

      if (objc != 4) {
        Tcl_WrongNumArgs(interp, 2, objv, "port subname");
        return TCL_ERROR;
      }
      port = Tcl_GetString(objv[2]);
      name = Tcl_GetString(objv[3]);

      for(p=0; p < h->info->nports; p++)
        if (!strcmp(h->info->ports[p].name, port)) break;
      if (p >= h->info->nports) {
        Tcl_AppendResult(interp, "unknown port ", port, NULL);
        return TCL_ERROR;
      }

      data = malloc(h->info->ports[p].data_size);
      if (!data) {
        Tcl_AppendResult(interp, "not enough memory", NULL);
        return TCL_ERROR;
      }
      gnmx_atexit_push(h->dlh);
      h->info->ports[p].init_data(data);
      if (h->info->ports[p].simple)
        s = h->info->ports[p].read.simple(h->c, data);
      else {
        if (name[0] == '\0') {
          Tcl_AppendResult(interp, "missing port set name", NULL);
          return TCL_ERROR;
        }
        s = h->info->ports[p].read.multiple(h->c, name, data);
      }
      gnmx_atexit_pop();
      if (s) {
        free(data);
        Tcl_AppendResult(interp, h->info->strerror(h->c), NULL);
        return TCL_ERROR;
      }
      s = h->info->ports[p].json_print(&json, data);
      free(data);
      if (s) {
        Tcl_AppendResult(interp, h->info->strerror(h->c), NULL);
        return TCL_ERROR;
      }

      Tcl_SetObjResult(interp, Tcl_NewStringObj(json, -1));
      free(json);
      break;
    }
  }

  return TCL_OK;
}


/* --- gnmx_requestcb ------------------------------------------------------ */

static void
gnmx_requestcb(genom_client c, int rqstid, int error, uint32_t report,
               void *output, void *data)
{
  struct gnmx_handle *h = data;
  const struct genom_service_info *r;
  static const int argc = 6;
  Tcl_Obj *argv[argc];
  char *json = NULL;
  int i, s;

  /* command */
  argv[0] = Tcl_NewStringObj("request::event", -1);
  argv[1] = Tcl_NewStringObj(h->info->instance(c), -1);
  argv[2] = Tcl_NewIntObj(rqstid);

  /* status */
  if (error == 0 && report == 0) {
    if (output == NULL)
      argv[3] = Tcl_NewStringObj("sent", -1);
    else
      argv[3] = Tcl_NewStringObj("done", -1);
  } else
    argv[3] = Tcl_NewStringObj("error", -1);

  /* error string, or empty */
  if (error || report)
    argv[4] = Tcl_NewStringObj(h->info->strerror(c), -1);
  else if (output) {
    r = h->info->service_info(c, rqstid);
    if (!r)
      argv[4] = Tcl_NewStringObj("internal error: rqstid with no data", -1);
    else {
      s = r->json_print(&json, output);
      if (s)
        argv[4] = Tcl_NewStringObj(strerror(s), -1);
      else
        argv[4] = Tcl_NewStringObj("", -1);
    }
  } else
    argv[4] = Tcl_NewStringObj("", -1);

  /* json output */
  if (json) {
    argv[5] = Tcl_NewStringObj(json, -1);
    free(json);
  } else
    argv[5] = Tcl_NewStringObj("", -1);

  /* invoke callback */
  for(i = argc-1; i >= 0; i--) Tcl_IncrRefCount(argv[i]);
  s = Tcl_EvalObjv(h->interp, argc, argv, TCL_EVAL_GLOBAL);
  for(i = argc-1; i >= 0; i--) Tcl_DecrRefCount(argv[i]);
  if (s != TCL_OK)
    Tcl_BackgroundError(h->interp);
}


/* --- gnmx_delete_cmd ----------------------------------------------------- */

static void
gnmx_delete_module_cmd(ClientData data)
{
  struct gnmx_handle *h = data;
  struct gnmx_handle *i;

  if (h) {
    if (h == gnmx_clients)
      gnmx_clients = h->next;
    else {
      for(i = gnmx_clients; i->next && i->next != h; i = i->next)
        /* empty body */;
      if (i->next == h) i->next = h->next;
    }

    gnmx_unwatch(h);
    if (h->c){
      gnmx_atexit_push(h->dlh);
      h->info->fini(h->c);
      gnmx_atexit_pop();
    }

    if (h->dlh) {
      gnmx_atexit_invoke(h->dlh);
      dlclose(h->dlh);
    }
    if (h->path) free(h->path);
    free(h);
  }
}
